/* Copyright 2006, 2007, 2009, 2010, 2018 Kevin Ryde.

   This file is part of pngtextadd.

   pngtextadd is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   pngtextadd is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with pngtextadd; see the file COPYING.  Failing that, goto
   <http://www.gnu.org/licenses/>.
*/

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <png.h>

#define VERSION 4

void
error (png_structp png_ptr, png_const_charp str)
{
  fprintf (stderr, "error: %s\n", str);
}

void
warning (png_structp png_ptr, png_const_charp str)
{
  fprintf (stderr, "warning: %s\n", str);
}

static const struct option longopts[] =
  {
    /* These options set a flag. */
    {"compress", no_argument,       NULL, 'c'},
    {"keyword",  required_argument, NULL, 'k'},
    {"text",     required_argument, NULL, 't'},
    {"lang",     required_argument, NULL, 'l'},
    {"lkeyword", required_argument, NULL, 'K'},
    {"verbose",  no_argument,       NULL, 'v' },
    {"version",  no_argument,       NULL, 'V' },
    {"help",     no_argument,       NULL, 'h'},
    {0, 0, 0, 0}
  };

int
main (int argc, char **argv)
{
  FILE *input_fp;
  char *filename;
  char *keyword = NULL;
  char *lkeyword = "";
  char *lang = NULL;
  char *text = NULL;
  png_infop info_ptr;
  int compression = PNG_TEXT_COMPRESSION_NONE;
  int itxt_compression = PNG_ITXT_COMPRESSION_NONE;
  int option_version = 0;
  int option_verbose = 0;

  {
    int index = 0;
    for (;;)
      {
        int c = getopt_long (argc, argv, "c", longopts, &index);
        switch (c) {
        case 'c':
          compression = PNG_TEXT_COMPRESSION_zTXt;
          itxt_compression = PNG_ITXT_COMPRESSION_zTXt;
          break;

        case 'h':
          printf ("pngtextadd --keyword AAA --text BBB [--options]\n");
          printf ("    --help            print this message\n");
#ifdef PNG_iTXt_SUPPORTED
          printf ("    --lang LLL\n");
          printf ("    --lkeyword LLL\n");
          printf ("iTXt supported\n");
#else
          printf ("iTXt not supported\n");
#endif
          printf ("Home page http://user42.tuxfamily.org/pngtextadd/index.html\n");
          exit (0);
          break;

        case 'k':
          keyword = optarg;
          break;

        case 'K':
#ifndef PNG_iTXt_SUPPORTED
          fprintf (stderr, "no iTXt in this libpng\n");
          exit (1);
#endif
          lkeyword = optarg;
          break;

        case 'l':
#ifndef PNG_iTXt_SUPPORTED
          fprintf (stderr, "no iTXt in this libpng\n");
          exit (1);
#endif
          lang = optarg;
          break;

        case 't':
          text = optarg;
          break;

        case 'v':
          option_verbose++;
          break;
        case 'V':
          option_version = 1;
          break;

        case -1:
          goto got_options;

        default:
          fprintf (stderr, "Unrecognised option: %c\n", c);
          exit (1);
          break;
        }
      }
  got_options:

    if (option_version)
      {
        printf ("pngtextadd version %d\n", VERSION);
        if (option_verbose)
          fputs (png_get_header_version(NULL), stdout);
        exit (0);
      }

    if (optind >= argc)
      {
        filename = "standard input";
        input_fp = stdin;
      }
    else if (optind + 1 < argc)
      {
        fprintf (stderr, "Only one filename can be given\n");
        exit (1);
      }
    else
      {
        filename = argv[optind];
        input_fp = fopen(filename, "rb");
        if (input_fp == NULL)
          {
            fprintf (stderr, "Cannot open %s\n", filename);
            exit (1);
          }
      }
  }

  if (keyword == NULL)
    {
      fprintf (stderr, "--keyword missing\n");
      exit (1);
    }

  if (text == NULL)
    {
      fprintf (stderr, "--text missing\n");
      exit (1);
    }

  {
    png_structp png_ptr;
    png_text *text_ptr, textnew;
    int i, num_text;

    png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING,
                                      NULL, error, warning);
    if (png_ptr == NULL)
      {
        fprintf (stderr, "Cannot create png structure\n");
        exit (1);
      }

    info_ptr = png_create_info_struct (png_ptr);
    if (info_ptr == NULL)
      {
        fprintf (stderr, "Cannot create png info structure\n");
        exit (1);
      }

    if (setjmp (png_jmpbuf (png_ptr)))
      {
      aborting:
        fprintf (stderr, "aborting\n");
        exit (1);
      }

    png_init_io(png_ptr, input_fp);

    png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);

    if (fclose (input_fp) != 0)
      {
        fprintf (stderr, "Error closing input\n");
        exit (1);
      }

    num_text = png_get_text (png_ptr, info_ptr, &text_ptr, NULL);

    for (i = 0; i < num_text; i++)
      {
        if (strcmp (text_ptr[i].key, keyword) == 0)
          {
            fprintf (stderr, "Replacing keyword \"%s\"\n", keyword);
            text_ptr[i].key = keyword;
            text_ptr[i].text = text;
            text_ptr[i].text_length = strlen (text);
            text_ptr[i].compression = compression;
#ifdef PNG_iTXt_SUPPORTED
            if (lang)
              {
                text_ptr[i].compression = itxt_compression;
                text_ptr[i].lang = lang;
                text_ptr[i].lang_key = lkeyword;
                text_ptr[i].itxt_length = strlen (text);
                text_ptr[i].text_length = 0;
              }
#endif
            goto munged;
          }
      }

    memset (&textnew, '\0', sizeof (textnew));
    textnew.key = keyword;
    textnew.text = text;
    textnew.text_length = strlen (text);
    textnew.compression = compression;
#ifdef PNG_iTXt_SUPPORTED
    if (lang)
      {
        textnew.compression = itxt_compression;
        textnew.lang = lang;
        textnew.lang_key = lkeyword;
        textnew.itxt_length = strlen (text);
        textnew.text_length = 0;
      }
#endif
    png_set_text (png_ptr, info_ptr, &textnew, 1);
  munged:
    ;
  }

  {
    FILE *output_fp;
    png_structp png_ptr;

    if (input_fp == stdin)
      {
        output_fp = stdout;
      }
    else
      {
        output_fp = fopen (filename, "wb");
        if (output_fp == NULL)
          {
            fprintf (stderr, "Cannot rewrite file\n");
            exit (1);
          }
      }

    png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING,
                                       NULL, error, warning);
    if (png_ptr == NULL)
      {
        fprintf (stderr, "Cannot create png structure\n");
        exit (1);
      }

    if (setjmp (png_jmpbuf (png_ptr)))
      goto aborting;

    png_init_io(png_ptr, output_fp);

    png_write_png (png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);

    if (fclose (output_fp) != 0)
      {
        fprintf (stderr, "Error closing output\n");
        exit (1);
      }
  }

  return 0;
}
